#include "QLoader.h"
#include "QFileElement.h"
#include "NodeCreationProxy.h"
#include "X3DAbstractNode.h"
#include "Scene.h"

#include <qxml.h>
#include <qfile.h>
#include <iostream>
#include <unistd.h>

using namespace X3DTK;
using namespace std;

namespace X3DTK {

class QSAX2X3DHandler : public QXmlDefaultHandler
{
public: 
  QSAX2X3DHandler(NodeCreationProxy *nodeCreationProxy);
  ~QSAX2X3DHandler();
  
  bool startDocument();
  bool startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts);
  bool endElement(const QString &namespaceURI, const QString &localName, const QString &qName);
  bool endDocument();
/*  void fatalError(const Q_CPP_NAMESPACE_QUALIFIER SAXParseException& exception);
  void error(const Q_CPP_NAMESPACE_QUALIFIER SAXParseException& exception);*/
  
  inline X3D::Scene *getScene() const {return _scene;};
  
private:
  MFAbstractNode _nodeStack;
  X3D::Scene *_scene;
  NodeCreationProxy *_ncp;
  bool _sceneDescription;
  unsigned int _depth;
};

}

X3DTK::QSAX2X3DHandler::QSAX2X3DHandler(NodeCreationProxy *nodeCreationProxy)
: _scene(0), _ncp(nodeCreationProxy), _sceneDescription(false), _depth(0)
{
}

X3DTK::QSAX2X3DHandler::~QSAX2X3DHandler()
{
}

bool X3DTK::QSAX2X3DHandler::startDocument()
{
  _scene = 0;
  _sceneDescription = false;
  _depth = 0;
  
  return true;
}

bool X3DTK::QSAX2X3DHandler::startElement(const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &atts)
{
  //_depth == 0, meaning that all the fathers of the node are defined nodes(have a Creation Function)
  //This variable is used to jump a sub-tree of the graph scene, when an unknown node is met.
  if (_depth == 0)
  {
    QFileElement e((const char *)qName, &atts);

    int index;  
    if (_sceneDescription)
    {
      SFAbstractNode Father = _nodeStack.back();
      SFAbstractNode N;
      index = e.getIndexAttribute("USE");
      
      if (index != -1)
      {
        N = X3DAbstractNode::DEFDict.getNodeOfName(e.getAttribute(index));
      }  
      else  
      {
        X3D::SFNode cN = _ncp->createFromName(e.getName());
        if (cN == 0)
        {
          _depth = 1;
          return false;
        }
        cN->load(&e);
        N = cN;
      }
      index = e.getIndexAttribute("DEF");
      if (index != -1)
        X3DAbstractNode::DEFDict.insert(e.getAttribute(index), N);

      Father->setChild(N);
      _nodeStack.push_back(N);  
    }  
  
    if (e.getName() == "Scene")
    {
      _sceneDescription = true;
      _scene = new X3D::Scene();
      _nodeStack.push_back(_scene);
    } 
  }
  else
    ++_depth;
    
  return true;  
}

bool X3DTK::QSAX2X3DHandler::endElement(const QString &namespaceURI, const QString &localName, const QString &qName)
{
  if (_sceneDescription)
  {
    if ((_depth == 0) && (_nodeStack.size() > 0))
      _nodeStack.pop_back();
    else
      --_depth;  
  }
  
  return true;
}

bool X3DTK::QSAX2X3DHandler::endDocument()
{
  X3DAbstractNode::DEFDict.makeDEFNonAmbiguous();
  _nodeStack.clear();

  return true;
}

/*void X3DTK::QSAX2X3DHandler::error(const SAXParseException& exception)
{
  char* message = XMLString::transcode(exception.getMessage());
  cout << "Error: " << message
       << " at line: " << exception.getLineNumber()
       << endl;
}

void X3DTK::QSAX2X3DHandler::fatalError(const SAXParseException& exception)
{
  char* message = XMLString::transcode(exception.getMessage());
  cout << "Fatal Error: " << message
       << " at line: " << exception.getLineNumber()
       << endl;
}
*/

class X3DTK::QLoaderImplementation
{     
public:
  QLoaderImplementation(){};
  
  QXmlSimpleReader *_parser;
  QSAX2X3DHandler *_handler;
};

QLoader::QLoader(NodeCreationProxy *nodeCreationProxy)
: _impl(new QLoaderImplementation())
{
  ncp = nodeCreationProxy;
  
  _impl->_parser = new QXmlSimpleReader();
  _impl->_handler = new QSAX2X3DHandler(nodeCreationProxy);
  _impl->_parser->setContentHandler(_impl->_handler);
}

QLoader::~QLoader()
{
  delete _impl->_parser;
  delete _impl->_handler;
}

X3D::Scene *QLoader::loadFile(const char *file, bool fileValidation) const 
{
  SFString sfile(file);
  
  // Goes to the directory of the file.
  char *cwd = (char *)malloc(65535);
  char *value = getcwd(cwd, 65535);
  if (value == 0)
  {
    cerr << "QLoader::load: unable to get current working directory!" << endl;
    return 0;
  }  
  
  if (chdir(sfile.getPath()) != 0)
  {
    cerr << "QLoader::load: unable to set the current working directory to !" << sfile.getPath()<< endl;
    return 0;
  }
  
  QFile xmlFile((const char *)sfile.getFile());
  QXmlInputSource source(&xmlFile);   
  
  try 
  {
    _impl->_parser->parse(source);
  }
  catch (...) 
  {
    cout << "Unexpected Exception \n" ;
  }
  
  // Goes back to the initial directory.
  if (chdir(cwd) != 0)
  {
    cerr << "QLoader::load: unable to set the current working directory to !" << cwd << endl;
    return 0;
  }
  
  free(cwd);
  
  X3D::Scene *scene = _impl->_handler->getScene();
  if (scene != 0)
    scene->setFileName(file);
      
  return scene;
}

